/*
 * Robot.java
 *
 * Created on July 27, 2001, 11:06 AM
 */

/**
 *
 * @author  Steve Perretta
 * @version
 *  
 */

/**
 *    Robot.java contains the source code used to control
 * the robot in both the simulator and the serial port programs.
 * The class defined here runs as an independent thread that will
 * allow you to continuously poll sensor data, and set the robot's
 * effectors accordingly. The interaction between this class and 
 * the rest of the program is facilitated via instances of the 3
 * following classes: CurrentRobotState, Sensor[], and Motor.
 * Once the local references to these objects have been initialized,
 * all data regarding the robot can be set or retrieved.
 * 
 * Here are the fields and methods you will need to access each
 * object:
 *
 * ------------------------------------------------------------------
 * CurrentRobotState:
 *     static fields: int ARM_UP, int ARM_DOWN, 
 *                    int GRIP_OPEN, int GRIP_CLOSED
 *     methods:
 *         >>> The following return or accept the above constants
 *         int     getArmState() 
 *         void    setArmState(int position)
 *         int     getGripperState()
 *         void    setGripperState(int position)
 *
 *         >>> The following are predicates
 *         boolean isObjectPresent()
 * ------------------------------------------------------------------
 *  
 * ------------------------------------------------------------------
 *  The reference to Sensor[] refers to an array of 8 Sensor objects.
 *  Sensor:
 *     methods:
 *         >>> The following return sensor values. Distance sensor
 *         >>> readings range from 0 to 1023; 0 - 5 generally means
 *         >>> no object is within the sensors range. Light sensor 
 *         >>> readings range around 5 to 512; readings between 495
 *         >>> and 512 correspond to "total darkness", while readings
 *         >>> below 50 correspond to maximum lighting.
 *         int getDistValue()
 *         int getLightValue()
 * ------------------------------------------------------------------
 *
 * ------------------------------------------------------------------
 * Motor:
 *     methods: 
 *         >>> The speed value for each motor can range from -9 to 9.
 *         >>> Wheel position values for each of the 2 wheels are 
 *         >>> incremented if the wheel speed is positive (and > 0)
 *         >>> and moving, decremented if the speed is negative and
 *         >>> moving. In other words, while the robot is stuck (e.g.
 *         >>> pushing against a wall) or if the motor speeds are 
 *         >>> set to 0, the position values will not change.
 *         void setMotorSpeeds(int leftSpeed, int rightSpeed) 
 *         long getLeftPosition() 
 *         long getRightPosition()
 * ------------------------------------------------------------------
 *
 * ------------------------------------------------------------------
 * MessagePasser:
 *         > This class enables the passing of String objects (i.e. 
 *         > messages, commands, etc.) to and from an instantiation of 
 *         > the Robot class. This allows you to interact with your 
 *         > controller when its being executed in the simulator, 
 *         > without having to mess around with standard I/O streams to 
 *         > and from a terminal window (although you can use this method 
 *         > if you like).
 *     methods:
 *         >>> accepts a valid String, which is then appended to the Output 
 *         >>> text area on the simulator GUI. Strings can be literals or
 *         >>> references to existing String objects. The String argument need 
 *         >>> not be terminated with a newline character. This method returns
 *         >>> immediately.
 *         void postMessage(String message)
 *
 *         >>> returns a String represention of the text entered in the Input 
 *         >>> text field on the simulator GUI. This value may be null if no 
 *         >>> text is entered. This method only returns when the Enter key is 
 *         >>> pressed while the Input text field has focus. In this way it 
 *         >>> behaves much like BufferedReader's readline() method, which
 *         >>> blocks until the user hits Enter.
 *         String getMessage() 
 * -------------------------------------------------------------------
 *
 *   The rest of the code in this file serves as an example control
 * program that illustrates the use of the interfaces outlined above.
 * Certain methods define the public interface of this class and should
 * not be altered. Any Robot.java file you create needs to have these
 * methods or it will not compile. Please read the comment block preceding
 * each method definition for a full description. It should also be noted
 * that the methods defined below are only examples; there are obviously
 * many different implementations that achieve the same effect. 
 *   As the control algorithms you implement get more complex, you may
 * wish to create your own classes. It is not recommended that you 
 * run any new objects as threads. Both the simulation and serial port
 * programs run multiple threads on top of the control thread defined 
 * here. Issues involving synchronization and priority will most likely
 * cause the program to become unstable, and exceptions thrown in 
 * response to this instability will terminate ALL threads in an 
 * "unclean" manner.
 */     

public class Robot extends Thread {
   /* Changing the following fields is not recommended.
    * The variable types should NOT be altered, but if
    * you decide to change the variable names be sure to
    * change all occurences throughout the code. 
    */
    private CurrentRobotState currentData;
    private MessagePasser     messagePasser;
    private Sensor[]          sensors; 
    private Motor             motorState;
    private int               gripperState;
    private int               armState;
    private boolean           objPresent;
    private boolean           active;

    // Add your own fields here....
    private int distThreshold = 500;

   
    /** Constructor -- DO NOT MODIFY 
     */
    public Robot(CurrentRobotState data, MessagePasser ma) {
        currentData = data;
        messagePasser = ma;
        active = true;
    } 

    /** This method should be called from the run() method - before
      * entering the while loop. These assignments cannot be made in the
      * constructor (?)
      */
    private void getRobotObjects() {
        sensors  = currentData.getSensorValues();
        motorState = currentData.getMotorState();
    }   
    
    /** Essentially a "wrapper" that sets the foreward speed 
     */
    private void setForwardSpeed() {
        motorState.setMotorSpeeds(3, 3);
    }
    
    /** Another "wrapper" method for setting turns
     */
    private void setTurnSpeed(int left, int right) {
        motorState.setMotorSpeeds(left, right);
    }
    
    /** This method simply reads the current distance sensor values
     *  from the side and front sensors (i.e. 0 through 5), and 
     *  returns the index of the "hottest" sensor. If all sensors are
     *  0, then -1 is returned. 
     */
    private int testProximity() {
        int max = 0;
        int closeIndex = -1;
        
        for (int i = 0; i < 6; i++) {
            int tempVal = sensors[i].getDistValue();
            if(tempVal > distThreshold) {
                if(tempVal > max) {
                    max = tempVal;
                    closeIndex = i;
                }
            }
        }
        return closeIndex;
    }
    
    /** This method defines a routine where the robot spins in place
     *  and polls the light sensors. When at least one sensor detects
     *  a significant amount of light, the robot will begin to move
     *  toward the light source. This is a very simple controller, and
     *  as you might expect, it will spin infinitely if no light is
     *  present or if the light is too far away to detect. Also note
     *  that MessagePasser's postMessage() method is demonstrated. 
     */
    private void aimAtLight() {
        int min = 500;               // dark 
        int closeIndex = -1;
        int lightThresholdMin = 450; // light is barely detectable
        int lightThresholdMax = 150; // light source is close
        boolean atLight = false;
   
        // Execute loop until we reach the light source
        while(!atLight) {
            for(int i = 0;i < 8;i++) {
                int sVal = sensors[i].getLightValue();
                if(sVal < lightThresholdMin) {
                    if(sVal < min) {
                        min = sVal;
                        closeIndex = i;
                    }
                }
            }
            // If we're at the light, then stop and return from
            // this method
            if(min < lightThresholdMax) {
                motorState.setMotorSpeeds(0,0);
                atLight = true;
            }
            switch(closeIndex) {
                // Light is detected to the left, so vear left
                case 0:
                case 1: motorState.setMotorSpeeds(-1, 3);
                        messagePasser.postMessage("vearing left");
                        break;
                // Light is detected to in front, so go straight
                case 2:
                case 3: motorState.setMotorSpeeds(3, 3);
                        messagePasser.postMessage("forward");
                        break;
                // Light is detected to the right, so vear right        
                case 4:
                case 5: motorState.setMotorSpeeds(3, -1);
                        messagePasser.postMessage("vearing right");
                        break;
                // Slowly spin and poll the light sensors
                default: motorState.setMotorSpeeds(-1, 1);
                         messagePasser.postMessage("polling");
            }
        }
    }
    
    /** This piece of control illustrates the use of method calls 
     *  which move the robot's arm and gripper. One of the hardest
     *  single tasks you will confront when writing controllers for
     *  the Khepera is determining what objects can be grabbed and 
     *  then picking them up. Randomly trying to grab something that
     *  is next to the robot (based on distance sensor readings) will
     *  very rarely work, and will greatly increase the chances of the
     *  robot getting stuck. The best strategy, in very general terms,
     *  is to: 1) somehow determine if some obstacle is "grippable",
     *  2) center on the object (i.e. position the robot so that its
     *  two front sensors are squared on the object), 3) adjust the 
     *  distance between the robot and the object so that when the 
     *  arm goes down, the front part of the arm (which ends up 
     *  protruding about .5 inches out in front of the REAL robot when the 
     *  arm is down) does not land on top of the object, 4) drop the arm 
     *  down and read the gripper sensor to determine if the object is
     *  between the grippers, 5) close the gripper and raise the arm.
     *  Something to consider during steps 4 and 5 is the time it
     *  takes to raise/lower the arm, and open/close the gripper.
     *  These actions typically have to be done in the proper order,
     *  and have to be given time in between each action to complete.
     *  In the following method you will see that sleep() has been used
     *  to ensure that, for example, the arm gets a chance to lower
     *  itself into the down position before the gripper closes, and 
     *  that time is also given to make sure the gripper is done closing 
     *  before the arm is raised again. 
     */ 
    private void grabObject() {
        /* Assumes that object has been found, and that we are in
         * a suitable orientation with respect to the object to
         * try a grab.
         */
        try {
            currentData.setArmState(CurrentRobotState.ARM_DOWN);
            Thread.sleep(1000L);
            if(currentData.isObjectPresent()) {
                currentData.setGripperState(CurrentRobotState.GRIP_CLOSED);
                Thread.sleep(1000L);
            }
            currentData.setArmState(CurrentRobotState.ARM_UP);
        } catch (InterruptedException e) {}
    }
    
    /** This method shows you how to poll the wheel positions to
     *  determine if the robot is stuck. As mentioned previously,
     *  these positions change as long as the wheels are actually
     *  rotating, so when the robot gets stuck we should see these
     *  values remain constant after successive reads. If you poll
     *  these values, you may have to sleep some small amount of time
     *  to ensure that the position values get updated. Although these
     *  values are updated quickly, it is possible that back to back
     *  requests for position values may retrieve the same data 
     *  eventhough the robot is moving. If you determine that the 
     *  robot is stuck, the only way to free it is to either go forward
     *  or backward depending on whether the robot has made contact
     *  with an obstacle within its front half or back half, 
     *  respectively. 
     */
    private boolean isRobotStuck() {
        long lPos = motorState.getLeftPosition();
        long rPos = motorState.getRightPosition();
        
        // Calls to sleep() can throw exception
        try {
            Thread.sleep(500L); 
        } catch (InterruptedException e) {}
        if(motorState.getLeftPosition() == lPos &&
           motorState.getRightPosition() == rPos)
              return true;
        return false;
    }
        
        
    
    /*  Terminates this thread. Called from the main program. 
     *    --- DO NOT REMOVE ---   
     */
    public void killControl() {
        active = false;
    }
    
    /*     This method defines the repetitive execution pattern
     *  of this thread. The call to getRobotObjects() must remain
     *  as the first statement in this method. This ensures that:
     *  1) the objects that define the robot are initialized before
     *  they are used, and 2) initializing them just prior to execution, 
     *  ensures that time has been given to create these objects within 
     *  the main program.
     */     
    public void run() {

        getRobotObjects();  // always call this !
        
        while(active) {
            // add method calls and other code here...
        }
    }
}


